program tbcnrand

!   This tests the sequential version of the BCN pseudorandom number generator.
!   David H Bailey    2008-08-25

!   The sequence generated is (x_k), where x_k = first 53 binary digits of the
!   binary expansion of alpha_{2,3} beginning at binary position 53*k + aa,
!   normalized by 2^53.  The parameter aa is the starting point of the sequence,
!   or in other words the seed of the pseudorandom sequence.  To obtain the 
!   maximum period, aa should be set to at least 3^33 = 5.5590606e15.  The value
!   of aa must not exceed 2^53 = 9.0071992e15.  When aa is set in this range, the
!   period of the generator is 2x3^32 = 3.7060404e15.  In general, the period
!   of the generator is 2x3^(p-1), where 3^p is the largest power of 3 less 
!   than aa.  However the generated sequence will not match the binary digits of
!   alpha_{2,3} if the range of sequence indices, namely [aa, aa+53*n], includes
!   a power of 3.

!   The bcnrand routine is designed for simple parallelization.  For example,
!   in an MPI program, suppose that kpr is the processor number and npr is the
!   number of processors.  Then the line 
!       call bcnrand (n/npr, aa + 53*n/npr*kpr, x)
!   generates on each processor a section of length n/npr.  In this way, the
!   npr processors collectively have the same n-long sequence (provided that
!   n is divisible by npr) as is generated on a single processor system by
!   means of the line
!       call bcnrand (n, aa, x)

!   On certain systems with a fused multiply-add instruction, see !> comments.

!   Parameters:
!     n    Length of the sequence desired.
!     np   Number of initial values that are output at start and end.
!     aa   Seed of sequence; default = 3^33 + 100; should satisfy
!          3^33 + 100 <= aa <= 2^53.

!   NOTE:  This includes a dummy "second" timing function below.  Replace
!   with a system-specific verison for nonzero output run times.

implicit none
integer i, n, np
parameter (n = 100000000, np = 10)
real*8 aa, x(n), tm0, tm1
parameter (aa = 3.d0**33 + 100.d0)

write (6, 1) n, aa
1 format ('bcnrand test program'/'n =',i10,' starting index =',f19.0)
tm0 = second ()
call bcnrand (n, aa, x)
tm1 = second ()
write (6, 2) tm1 - tm0, np, (x(i), i = 1, np)
2 format ('CPU time =',f16.6/'first',i8,' elements of sequence ='/(4f19.15))
write (6, 3) np, (x(i), i = n - np + 1, n)
3 format ('final',i8,' elements of sequence ='/(4f19.15))
stop
end
