program tbcnrandx

!   This tests the sequential version of the BCN pseudorandom number generator.
!   David H Bailey    2008-08-25

!   The sequence generated is (x_k), where x_k = first 53 binary digits of the
!   binary expansion of alpha_{2,3} beginning at binary position 53*k + aa,
!   normalized by 2^53.  The parameter aa is the starting point of the sequence,
!   or in other words the seed of the pseudorandom sequence.  To obtain the 
!   maximum period, aa should be set to at least 3^33 = 5.5590606e15.  The value
!   of aa must not exceed 2^53 = 9.0071992e15.  When aa is set in this range, the
!   period of the generator is 2x3^32 = 3.7060404e15.  In general, the period
!   of the generator is 2x3^(p-1), where 3^p is the largest power of 3 less 
!   than aa.  However the generated sequence will not match the binary digits of
!   alpha_{2,3} if the range of sequence indices, namely [aa, aa+53*n], includes
!   a power of 3.

!   The bcnrand routine is designed for simple parallelization.  For example,
!   in an MPI program, suppose that kpr is the processor number and npr is the
!   number of processors.  Then setting the seed on processor kpr to be
!   aa + 53*n/npr*kpr, where aa is the seed in the single-CPU case, generates
!   sequences of length n / npr on each processor, which collectively is the
!   same as the sequence generated on a single CPU.

!   See comments below for complete details of using bcnrandx.
!   On certain systems with a fused multiply-add instruction, see !> comments.

!   Parameters:
!     n    Length of the sequence desired.
!     np   Number of initial values that are output at start and end.
!     aa   Seed of sequence; default = 3^33 + 100; should satisfy
!          3^33 + 100 <= aa <= 2^53.

!   NOTE:  This includes a dummy "second" timing function below.  Replace
!   with a system-specific verison for nonzero output run times.

implicit none
integer i, n, np2
parameter (np2 = 100, n = 100000000)
real*8 aa, bcnrandx, d1, d2, d3, expm2, p3i, r3i, t1, t2, t3, &
  dd1(2), dd2(2), dd3(2), tm0, tm1
parameter (p3i = 3.d0 ** 33, r3i = 1.d0 / p3i, aa = p3i + 100.d0)
external bcnrandx, expm2

write (6, 1) n, aa
1 format ('bcnrandx test program'/'n =',i10,' starting index =',f19.0)

!   Set the initial seed based on the position aa.  This code is only used
!   to match that of the BCN sequence beginning at position aa.

d2 = expm2 (aa - p3i, p3i)
d3 = aint (0.5d0 * p3i)
call ddmuldd (d2, d3, dd1)
d1 = aint (r3i * dd1(1))
call ddmuldd (d1, p3i, dd2)
call ddsub (dd1, dd2, dd3)
d1 = dd3(1)
if (d1 < 0.d0) d1 = d1 + p3i

!   Compute the first n3 elements of the sequence, one by one.

tm0 = second ()

do i = 1, n
  t1 = bcnrandx (d1)
enddo

tm1 = second ()
write (6, 2) tm1 - tm0, t1
2 format ('CPU time =',f16.6/'final value =',f19.15)

!   Now compute the final member of the sequence directly via the
!   exponential scheme as a validity check.

d2 = expm2 (aa - p3i + 53.d0*n, p3i)
d3 = aint (0.5d0 * p3i)
call ddmuldd (d2, d3, dd1)
d1 = aint (r3i * dd1(1))
call ddmuldd (d1, p3i, dd2)
call ddsub (dd1, dd2, dd3)
d1 = dd3(1)
if (d1 < 0.d0) d1 = d1 + p3i
t1 = r3i * d1
write (6, 3) t1
3 format ('check value =',f19.15)

stop
end
